@@ -11,7 +11,7 @@ import ipywidgets
 from IPython.display import display,clear_output
 
 def makeQRxls(fn, n=10):
-    qdict = {"QR codes" : [uuid.uuid4() for _ in  range(n)] }
+    qdict = {"UIDs" : [uuid.uuid4() for _ in  range(n)] }
     df = pd.DataFrame.from_dict(qdict)
     df.to_excel(fn, index=False)
     
@@ -49,10 +49,12 @@ def validate_sample_list(xls_fn,
         otherwise generate the info need for sample mixing/transfer: sample dictionary and list of mixing ops
     """
 
+    msg = []
+    
     xl = pd.ExcelFile(xls_fn)
     if sh_name is None:
         sh_name = xl.sheet_names[0]
-    print(f"reading data from {sh_name} ...") 
+    msg.append(f"reading data from {sh_name} ...") 
     df = xl.parse(sh_name, dtype={'Volume (uL)': float})    
         
     # get all samples
@@ -74,7 +76,7 @@ def validate_sample_list(xls_fn,
     # check whether every sample has a buffer and in the same row
     for sn in list(set(all_samples)-set(all_buffers)):
         if not isinstance(sdict[sn]["Buffer"], str): # nan
-            print(f"Warning: {sn} does not have a corresponding sample/buffer.")
+            msg.append(f"Warning: {sn} does not have a corresponding sample/buffer.")
         else:
             swell = sdict[sn]["Well"]
             bwell = sdict[sdict[sn]["Buffer"]]["Well"]
@@ -120,18 +122,20 @@ def validate_sample_list(xls_fn,
 
     rlen = np.asarray([len(rdict[rn]) for rn in rdict.keys()])
     if len(rlen[rlen<5])>0:
-        print(f"Consider consolidating the rows, more than two rows are half full.")
+        msg.append(f"Consider consolidating the rows, more than two rows are half full.")
 
     if generate_barcode:
         bcode,bcode_img = make_barcode(proposal_id, SAF_id, plate_id, dirname(xls_fn))
-        plt.imshow(bcode_img)
+        msg.append(f"writing barcode to {bcode}.pdf ...")
+        #plt.imshow(bcode_img)
+        display(bcode_img)
         t=plt.axis(False)
         if sh_name!=bcode:
-            print(f"Renaming the sheet name to {bcode} ...")
+            msg.append(f"Renaming the sheet name to {bcode} ...")
             ss = openpyxl.load_workbook(xls_fn)
             ss[sh_name].title = bcode
             ss.save(xls_fn)
-        return #bcode
+        return msg
     else:
         bcode = sh_name
         try:
@@ -153,6 +157,7 @@ def process_sample_lists(xls_fns, process_all_tabs=False, b_lim=4):
         ot2_layout should describe the labware present on the deck, with identifying bar/QR codes
     """
     pdict = {}
+    bdict0 = None
     if not isinstance(xls_fns, list):
         xls_fns = [xls_fns]
     for fn in xls_fns:
@@ -163,6 +168,10 @@ def process_sample_lists(xls_fns, process_all_tabs=False, b_lim=4):
             sh_list = [xl.sheet_names[0]]
         for sh_name in sh_list:
             bdict,wdict,mdict = validate_sample_list(fn, b_lim=b_lim, generate_barcode=False)
+            if bdict0 is None:
+                bdict0 = bdict
+            elif bdict['proposal_id']!=bdict0['proposal_id'] or bdict['proposal_id']!=bdict0['proposal_id']:
+                raise Exception("Cannot process plates from differernt proposal/SAF together.")
             pdict[sh_name] = [wdict, mdict]
     
     # get ot2 layout info
@@ -172,6 +181,7 @@ def process_sample_lists(xls_fns, process_all_tabs=False, b_lim=4):
     slist = []
     for bcode in pdict.keys():
         pname = bcode # labware name for the plate
+        plabel = bcode.split("-")[-1]
         row_name = None
         row_count = 0
         wdict,mdict = pdict[bcode]
@@ -182,7 +192,7 @@ def process_sample_lists(xls_fns, process_all_tabs=False, b_lim=4):
                 row_count += 1
                 if row_count%2:
                     # get labware name for the holder
-                    hname = f"h{int((row_count+1)/2):02d}"
+                    hname = f"{plabel}_h{int((row_count+1)/2):02d}"
             col = int(wn[1:])
             hpos = 2*col - (row_count%2) 
             
@@ -222,9 +232,9 @@ def read_OT2_layout(plate_slots, holder_slots):
 def generate_measurement_spreadsheet(fn, slist, holders, bdict):
     """ slist is generated by process_sample_lists()
         
-        holders is a dictionary of {holder_name: QR code}, derived values retuened 
+        holders is a dictionary of {holder_name: UID}, derived values returned 
         from read_OT2_layout(). this information needs to be saved in the spreadsheet, 
-        on the "QR codes" tab, so that the holder name can be found based on the QR 
+        on the "UIDs" tab, so that the holder name can be found based on the QR 
         code during the measurements
     """
     tslist = np.asarray(slist).T
@@ -238,14 +248,14 @@ def generate_measurement_spreadsheet(fn, slist, holders, bdict):
 
     cdict = {}
     cdict["holderName"] = holders.keys()
-    cdict["QR code"] = holders.values()
+    cdict["UID"] = holders.values()
     df2 = pd.DataFrame.from_dict(cdict)
 
     tname = f"{bdict['proposal_id']}-{bdict['SAF_id']}"
     
     with pd.ExcelWriter(fn) as writer:
         df1.to_excel(writer, index=False, sheet_name=tname)
-        df2.to_excel(writer, index=False, sheet_name="QR codes")
+        df2.to_excel(writer, index=False, sheet_name="UIDs")
             
 
 def generate_docs(ot2_layout, xls_fns,    
@@ -253,7 +263,8 @@ def generate_docs(ot2_layout, xls_fns,
                   b_lim=4,
                   plate_type = "corning_96_wellplate_360ul_flat",
                   holder_type = "lix_3x_holder_c",
-                  tip_type = "opentrons_96_tiprack_300ul"
+                  tip_type = "opentrons_96_tiprack_300ul",
+                  flow_rate_aspirate = 50, flow_rate_dispense = 50
                  ):
     """ ot2_layout should be a dictionary:
             {"plates" : "1,2",
@@ -298,6 +309,8 @@ def generate_docs(ot2_layout, xls_fns,
         protocol.append(f"    lbw{slot} = ctx.load_labware('{tip_type}', '{slot}')\n")
         tips.append(f"lbw{slot}")
     protocol.append(f"    pipet = ctx.load_instrument('p300_single', 'left', tip_racks=[{','.join(tips)}])\n")
+    protocol.append(f"    pipet.flow_rate.aspirate = {flow_rate_aspirate}\n")
+    protocol.append(f"    pipet.flow_rate.dispense = {flow_rate_dispense}\n")
 
     for st in transfer_list:
         src,sw,dest,dw,vol = st
@@ -328,6 +341,109 @@ def generate_docs(ot2_layout, xls_fns,
     
     print("Done.")
     
+def generate_docs2(ot2_layout, xls_fns,    
+                   run_name="test",
+                   b_lim=4,
+                   plate_types = ["corning_96_wellplate_360ul_flat", 
+                                  "biorad_96_wellplate_200ul_pcr"],
+                   holder_types = ["lix_3x_holder_c"],
+                   tip_types = ["opentrons_96_tiprack_300ul",
+                                "opentrons_96_tiprack_20ul"],
+                   pipets = {"left": {"type": "p300_single", "tip_size": "300ul", "maxV": 300}, 
+                             "right": {"type": "p20_single", "tip_size": "20ul", "maxV": 20}},
+                   flow_rate_aspirate = 0.3, flow_rate_dispense = 0.3  # fraction of the maxV
+                  ):
+    """ ot2_layout should be a dictionary, slot #: labware type
+            {"1" : "lix_3x_holder_c",
+             "2" : "corning_96_wellplate_360ul_flat",
+             "3" : "opentrons_96_tiprack_300ul"
+             "4" : "lix_3x_holder_c",
+             "6" : "opentrons_96_tiprack_20ul"}
+    """
+    print("Processing sample list(s) ...")
+    slist,transfer_list,bdict = process_sample_lists(xls_fns, b_lim=b_lim)
+    
+    h_slots = [k for k,l in ot2_layout.items() if l in holder_types]
+    p_slots = [k for k,l in ot2_layout.items() if l in plate_types]
+    t_slots = [k for k,l in ot2_layout.items() if l in tip_types]
+                    
+    print("Reading bar/QR codes, this might take a while ...")
+    ldict = read_OT2_layout(",".join(p_slots), ",".join(h_slots))
+    print(ldict)
+    
+    holders = {}
+    holder_qr_codes = chain(ldict['holders'].keys())
+    print(f"{len(ldict['holders'])} holders are available.")
+    for st in slist:
+        if not st[0] in holders.keys():
+            try:
+                holders[st[0]]= next(holder_qr_codes)
+            except StopIteration:
+                print("Error: Not enough sample holders for transfer.")
+                raise
+    print(f"{len(holders)} holders are needed.")
+
+    fn = f"{run_name}_protocol.py"
+    print(f"Generating protocol ({fn}) ...")
+    protocol = ["metadata = {'protocolName': 'sample transfer',\n",
+                "            'author': 'LiX',\n",
+                "            'description': 'auto-generated',\n",
+                "            'apiLevel': '2.3'\n",
+                "           }\n", 
+                "\n",
+                "def run(ctx):\n",]
+
+    for slot in p_slots+h_slots+t_slots:
+        protocol.append(f"    lbw{slot} = ctx.load_labware('{ot2_layout[slot]}', '{slot}')\n")
+
+    for k,p in pipets.items():
+        tips = ','.join([f"lbw{s}" for s in t_slots if p["tip_size"] in ot2_layout[s]])
+        protocol.append(f"    pipet_{k} = ctx.load_instrument('p300_single', 'left', tip_racks=[{tips}])\n")
+        protocol.append(f"    pipet_{k}.flow_rate.aspirate = {flow_rate_aspirate*p['maxV']}\n")
+        protocol.append(f"    pipet_{k}.flow_rate.dispense = {flow_rate_dispense*p['maxV']}\n")
+
+    # sorted by maxV, low to high
+    pvdict = {pipets[pn]["maxV"]:pn for pn in pipets.keys()}
+    pvdict = {k:pvdict[k] for k in sorted(pvdict.keys())}
+    vlist = list(pvdict.keys())
+    def select_pipet(v):
+        if (v>vlist).all():
+            raise Exception(f"requested transfer volume exceeds tip maximum")
+        elif (v<vlist).all():
+            p = pvdict[vlist[0]]
+        else:
+            p = pvdict[vlist[-1]]
+        return f"pipet_{p}"         
+    
+    for st in transfer_list:
+        src,sw,dest,dw,vol = st
+        if dest in holders.keys():
+            dest = holders[dest]
+        if src in ldict["plates"].keys():
+            sname = f"lbw{ldict['plates'][src]['slot']}.well('{sw}')"
+        elif src in holders.values():
+            sname = f"lbw{ldict['holders'][src]['slot']}.well('{ldict['holders'][src]['holder']}{sw}')"
+        else:
+            raise Exception(f"Unknown labware encountered: {src}")
+        if dest in ldict["plates"].keys():
+            dname = f"lbw{ldict['plates'][dest]['slot']}.well('{dw}')"
+        elif dest in holders.values():
+            dname = f"lbw{ldict['holders'][dest]['slot']}.well('{ldict['holders'][dest]['holder']}{dw}')"
+        else:
+            raise Exception(f"Unknown labware encountered: {dest}")
+        
+        protocol.append(f"    {select_pipet(vol)}.transfer({vol}, {sname}, {dname})\n")
+
+    fd = open(fn, "w+")
+    fd.writelines(protocol)
+    fd.close()
+    
+    fn = f"{run_name}.xlsx"
+    print(f"Writing measurement sequence to {fn}.")
+    generate_measurement_spreadsheet(fn, slist, holders, bdict)
+    
+    print("Done.")
+    
     
 def validatePlateSampleListGUI():
     propTx = ipywidgets.Text(value='',
@@ -360,10 +476,11 @@ def validatePlateSampleListGUI():
             outTxt.value = "upload the sample list spreadsheet first ..."
             return
         try:
-            validate_sample_list(flist[0], generate_barcode=True, 
+            msg = validate_sample_list(flist[0], generate_barcode=True, 
                                  proposal_id=propTx.value, 
                                  SAF_id=safTx.value, 
                                  plate_id=plateTx.value)
+            outTxt.value = "\n".join(msg)
         except Exception as e:
             s,r = getattr(e, 'message', str(e)), getattr(e, 'message', repr(e))
             outTxt.value = "Error: "+s
@@ -437,7 +554,7 @@ def checkHolderSpreadsheet(spreadSheet, sheet_name=0,
             if not (isinstance(e, int) or isinstance(e, float)):
                 raise Exception(f"non-numerical value in {f}: {e}")
             if e<=0 or np.isnan(e):
-                raise Exception(f"invalid value in {f}: {e}, possitive value required.")
+                raise Exception(f"invalid value in {f}: {e}, positive value required.")
     if 'volume' in allFields:
         if np.min(list(d['volume'].values()))<min_load_volume:
             raise Exception(f"load volume must be greater than {min_load_volume} ul!")
@@ -449,27 +566,38 @@ def checkHolderSpreadsheet(spreadSheet, sheet_name=0,
     if sp.min()<1:
         raise Exception(f"invalid sample positionL {sp.min()}.")
 
-    # check for duplicate sample name
-    sl = list(d['sampleName'].values())
-    for ss in sl:
-        if not check_sample_name(ss, check_for_duplicate=False):
-            raise Exception(f"invalid sample name: {ss}")
-        if sl.count(ss)>1 and str(ss)!='nan':
-            idx = list(d['holderName'])
-            hl = [d['holderName'][idx[i]] for i in range(len(d['holderName'])) if d['sampleName'][idx[i]]==ss]
-            for hh in hl:
-                if hl.count(hh)>1:
-                    raise Exception(f'duplicate sample name: {ss} in holder {hh}')
-    # check for duplicate sample position within a holder
-    hlist = list(np.unique(list(d['holderName'].values())))
-    idx = list(d['holderName'])
-    for hn in hlist:
-        plist = [d['position'][idx[i]] for i in range(len(d['holderName'])) if d['holderName'][idx[i]]==hn]
-        for pv in plist:
-            if plist.count(pv)>1:
-                raise Exception(f"duplicate sample position: {pv}")
-            
-    return hlist
+    sdict = {}
+    for (hn,pos,sn,bn) in zip(d['holderName'].values(), 
+                              d['position'].values(), 
+                              d['sampleName'].values(), 
+                              d['bufferName'].values()):
+        if not hn in sdict.keys():
+            sdict[hn] = {}
+        if str(sn)=='nan':
+            continue
+        if pos in sdict[hn].keys():
+            raise Exception(f"duplicate sample position {pos} in {hn}")
+        if not check_sample_name(sn, check_for_duplicate=False):
+            raise Exception(f"invalid sample name: {sn} in holder {hn}")
+        sdict[hn][pos] = {'sample': sn}
+        if str(bn)!='nan':
+            sdict[hn][pos]['buffer'] = bn 
+
+    for hn,sd in sdict.items():
+        plist = list(sd.keys())
+        slist = [t['sample'] for t in sd.values()]
+        for pos,t in sd.items():
+            if slist.count(t['sample'])>1:
+                raise Exception(f"duplicate sample name {t['sample']} in {hn}")
+            if not 'buffer' in t.keys():
+                continue
+            if not t['buffer'] in slist:
+                raise Exception(f"{t['buffer']} is not a valid buffer in {hn}")
+            bpos = plist[slist.index(t['buffer'])]
+            if (bpos-pos)%2:
+                raise Exception(f"{t['sample']} and its buffer not in the same row in holder {hn}")
+                    
+    return sdict
 
 def autofillSpreadsheet(d, fields=['holderName', 'volume']):
     """ if the filed in one of the autofill_fileds is empty, duplicate the value from the previous row
@@ -491,18 +619,24 @@ def autofillSpreadsheet(d, fields=['holderName', 'volume']):
 def validateHolderSpreadsheet(fn, proposal_id, SAF_id):
     # meant to be used by the users to attach to SAF
     # limit to 3 sample holders per spreadsheet
-    # validate sample list on the Holders tab and generate QR code for each holder 
+    # validate sample list on the Holders tab and generate UIDs for each holder 
     # beamline prints the QR codes and ship the holders to user
     print("Checking spreadsheet format ...")
-    hlist = checkHolderSpreadsheet(fn) 
+    sdict = checkHolderSpreadsheet(fn)
+    hlist = list(sdict.keys()) 
     if len(hlist)>3:
         raise Exception(f"Found {len(hlist)} sample holders. Only 3 are allowed.")
-
-    print("Generating QR codes ...")
+    ll = np.asarray([len(h) for h in hlist])
+    if (ll>5).any():  # for the purpose of fitting the text on the QR code stciker
+        raise Exception(f"Please limit the length of the holder names to 5 characters.")
+        
+    print("Generating UUIDs ...")
     wb = openpyxl.load_workbook(fn)
     wb[wb.sheetnames[0]].title = f"{proposal_id}-{SAF_id}"
-    ws1 = wb.create_sheet("QR codes")
-    ws1.append(["holderName", "QR code"])
+    if "UIDs" in wb.sheetnames:
+        del wb["UIDs"]
+    ws1 = wb.create_sheet("UIDs")
+    ws1.append(["holderName", "UID"])
     for i in range(len(hlist)):
         ws1.append([hlist[i], str(uuid.uuid4())])
     wb.save(fn)
